# copyright: Huynh Duong Phuong Vi
# www.sketchupforyou.com
module S4U
	module S4u_Linetoface
		class LineToFace
			def initialize()
				ext=IS_OSX ? ".pdf" : ".svg"
				cursor_path= File.join(File.dirname(__FILE__),("icon_mouse" + ext))
				@cursor_id = UI.create_cursor(cursor_path,0,0) if cursor_path
				@strings=S4U::S4u_Linetoface::Strings
				@scale_factor= Sketchup.version.to_i > 17 ? UI.scale_factor : 1
				@snap=20*@scale_factor
				@list_type=[@strings["Select 2 Points"],@strings["Select Height"]]
				@list_edges=[]
				@list_faces=[]
				@list_curves=[]
				@line_style=0
				@make_group=0
				@setting=S4U::S4u_Linetoface.instance_variable_get(:@setting)
				@make_group,@line_style=@setting
				@last_tool=S4U::S4u_Linetoface.instance_variable_get(:@last_tool)
			end
			def activate
				@edit_transform= axes_edit_transform
				@ip1 = Sketchup::InputPoint.new
				@ip2 = Sketchup::InputPoint.new
				@ip = Sketchup::InputPoint.new
				model=Sketchup.active_model
				model.active_view.lock_inference
				@sels=model.selection.to_a
				entities=model.active_entities
				@faces=@sels.select{|e| e && e.valid? && e.is_a?(Sketchup::Face) }
				@edge_faces=@faces.flat_map{|f| f.edges}.uniq

				edges=@sels.select{|e| e && e.valid? && e.is_a?(Sketchup::Edge) }.uniq
				#edges=@sels.select{|e| e && e.valid? && e.is_a?(Sketchup::Edge) && (e.faces & @sels.to_a).length==0}.uniq
				edge_curves=edges.select{|e| e.curve }
				@edges=edges-@edge_faces#-edge_curves
				check_hide_edge=!Sketchup.active_model.rendering_options['DrawHidden']
				if check_hide_edge
					@edges.select!{|e| e && e.valid? && !e.soft? && !e.hidden? }
					@edge_faces.select!{|e| e && e.valid? && !e.soft? && !e.hidden? }
					edge_curves.select!{|e| e && e.valid? && !e.soft? && !e.hidden? }
				end
				@curves=(edge_curves+@edge_faces).collect{|e| e.curve}.uniq
				if @edges.empty? && @faces.empty? #&& @curves.empty?
					UI.messagebox @strings["Select Edges or Faces"]
					model.select_tool nil
				end
				@hash_face_of_edges={}
				fs = @edges.flat_map{|p| p.faces }
				if fs[0]
					fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
					fs=fs.sort_by{|f| f[1]}.reverse
					fs.each{|f| @hash_face_of_edges[f[0]]=f[1] }
				end
				@list_continous_edges=[]
				if Sketchup.version.to_f>20
					Sketchup.active_model.start_operation TOOL
						curve=Sketchup.active_model.active_entities.weld(@edges)
						@list_continous_edges << curve.collect{|e| e.vertices}
					Sketchup.active_model.abort_operation
					edge_curves=edges.select{|e| e.curve }
					if check_hide_edge
						edge_curves.select!{|e| e && e.valid? && !e.soft? && !e.hidden? }
					end
					@curves=(edge_curves+@edge_faces).collect{|e| e.curve}.uniq
					@list_curves= get_point_from_curve(@curves,[@ip1.position,@ip2.position])
				else
					@list_continous_edges= S4U::S4u_Linetoface.find_weld(@edges,check_hide_edge)
				end
				# @hash_face_of_curves={}
				# fs = @curves.flat_map{|c| c.edges.flat_map{|p| p.faces} }
				# if fs[0]
					# fs=fs.sort_by{|f| f.object_id}.chunk{|f| f}.collect{|f| [f[0],f[1].length]}
					# fs=fs.sort_by{|f| f[1]}.reverse
					# fs.each{|f| @hash_face_of_curves[f[0]]=f[1] }
				# end
				# p "edges",@edges.length
				# p "curve" ,@curves.length
				# p "face",@faces.length
				# @faces_vertices=@faces.collect{|e| e.vertices}.flatten.uniq
				# @edges_vertices=@edges.collect{|e| e.vertices}.flatten.uniq
				# @curves_vertices==@curves.collect{|e| e.vertices}.flatten.uniq
				self.reset
			end
			def axes_edit_transform
				return Sketchup.active_model.axes.transformation
			end
			def onSetCursor
				UI.set_cursor(@cursor_id) if @cursor_id
			end
			def getExtents
				bb = Sketchup.active_model.bounds
				bb.add @ip1.position if @ip1.valid?
				bb.add @ip2.position if @ip2.valid?
				return bb
			end
			def getMenu(menu)
				menus=[]
				menus<< menu.add_item(@list_type[0]){ @line_style=0;set_status_text }
				menus<< menu.add_item(@list_type[1]){ @line_style=1;set_status_text }
				menus.each{|m| menu.set_validation_proc(m) { MF_UNCHECKED }}
				menu.set_validation_proc(menus[@line_style]) { MF_CHECKED }
				menu.add_separator
				menus<< menu.add_item(@strings["Group Faces"]){ @make_group=1-@make_group;set_status_text }
				menu.set_validation_proc(menus[-1]) { @make_group==0 ? MF_UNCHECKED : MF_CHECKED }
			end
			def onLButtonDoubleClick(flags, x, y, view)
				if @last_tool
					create_geometry(@last_tool)
					set_status_text(view)
					view.model.select_tool nil
				end
			end
			def deactivate(view)
				if @cancel
					@group.explode if @group && @group.valid? && @group.is_a?(Sketchup::Group) && (@make_group==0)
					@cancel.commit_operation
				end
				S4U::S4u_Linetoface.instance_variable_set(:@last_tool,@last_tool)
				@setting=[@make_group,@line_style]
				S4U::S4u_Linetoface.instance_variable_set(:@setting,@setting)
			end
			def get_vertices(face)
				return nil if not( face && (face.is_a?Sketchup::Face))
				ver=[]
				vertices=[]
				vertices =face.outer_loop.vertices*3
				(0..face.outer_loop.vertices.length-1).each { |i|
					j=face.outer_loop.vertices.length+i
					p1=vertices[j-1].position.clone
					p2=vertices[j].position.clone
					p3=vertices[j+1].position.clone
					v1=p1.vector_to p2
					v2=p2.vector_to p3
					ver.push vertices[j] if not( v1.samedirection? v2)
				}
				return ver
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@snap) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip="       " + @strings.GetString("1/3")
					return ip
				end
				return nil
			end
			def center_face(face,num=200)
				return nil if face.outer_loop.vertices.length> num #|| face.loops.length>1
				vertices=get_vertices(face)
				return nil if !vertices
				numver=vertices.length
				cp=Geom::Point3d.new
				vertices.each{|v|
					cp.x=cp.x+v.position.x
					cp.y=cp.y+v.position.y
					cp.z=cp.z+v.position.z
				}
				cp.x=cp.x/ numver
				cp.y=cp.y/ numver
				cp.z=cp.z/ numver
				return cp
			end
			def set_pointonedge(x,y,view,ip,edge)
				p1=edge.start.position.clone
				p2=edge.end.position.clone
				return nil if edge.length < (view.pixels_to_model(5*@snap, p1))
				points= [Geom.linear_combination(2.0/3, p1,1.0/3, p2),Geom.linear_combination(1.0/3, p1,2.0/3, p2)]
				ph = view.pick_helper
				ph.init(x, y)
				picked = points.find{|p| ph.test_point((p.transform ip.transformation),x,y,@snap) }
				if picked
					ip1=Sketchup::InputPoint.new(picked.transform ip.transformation)
					ip.copy! ip1
					view.tooltip= @strings.GetString("1/3")
					return ip
				end
				return nil
			end
			def get_bounds(element)
				b=nil
				if element.is_a?Sketchup::ComponentInstance
					b=element.definition.bounds
				elsif element.is_a?Sketchup::Group
					b=element.local_bounds
				elsif element.is_a?Sketchup::Image
					b=element.bounds
				else
					return nil
				end
				return (0..7).collect{|i| b.corner(i).clone.transform element.transformation } if b
			end
			def set_special_point(x,y,view,ip)
				ph = view.pick_helper
				ph.init(x, y)
				ph.do_pick(x, y)
				sels=view.model.selection.select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) || e.is_a?(Sketchup::Image))}
				points=[]
				@picked_element << ph.best_picked  if (e=ph.best_picked) && (e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)))
				@picked_element.select!{|e| e && e.valid?}
				sels+=@picked_element
				best_entity = sels.sort_by{|e| view.screen_coords(e.bounds.center).distance view.screen_coords(ip.position) }
				best_entity=best_entity[0..7].select{|e| e.valid? && (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) )}
				if !best_entity.empty?
					best_entity.each{|e|
						ps = get_bounds(e)
						points+=ps
						ps.permutation(2).to_a.each{|a|  points << Geom.linear_combination(0.5, a[0],0.5,a[1]) }
						# get_lines_bounds(e).each{|a|  points << Geom.linear_combination(0.5, a[0],0.5,a[1]) }
					}
				end
				points.uniq!
				picked_bounds = points.find_all{|p| ph.test_point(p,x,y,1.5*@snap) }
				if picked_bounds
					picked_bounds.each{|picked|
						model=view.model
						ray=view.pickray x, y
						if ray
							ray=[picked,ray[1].reverse!]
							item=model.raytest(ray,false)
							if !item
								@edge=true
								ip1=Sketchup::InputPoint.new(picked)
								ip.copy! ip1
								view.tooltip=@strings.GetString("BoundingBox")
								return picked
							end
						end
					}
				end
				if ip && ip.valid? && ip.face
					cf=center_face(ip.face)
					return ip if !cf
					cf.transform! ip.transformation
					ph = view.pick_helper
					ph.init(x, y)
					picked = ph.test_point(cf,x,y,@snap)
					if picked
						ip1=Sketchup::InputPoint.new(cf)
						@face=true
						view.tooltip=@strings.GetString("Center face")
						ip.copy! ip1
					else
						@edge=ip.face.edges.find{|e| set_pointonedge(x,y,view,ip,e) }
					end
				elsif ip.edge
					@edge=set_pointonedge(x,y,view,ip,ip.edge)
				end
				return ip
			end
			def onMouseMove(flags, x, y, view)
				@shift_press= (CONSTRAIN_MODIFIER_MASK&flags==CONSTRAIN_MODIFIER_MASK)
				@ctrl_press= (COPY_MODIFIER_MASK&flags==COPY_MODIFIER_MASK)
				@alt_press= (ALT_MODIFIER_MASK&flags==ALT_MODIFIER_MASK)
				@face=@edge=nil
				@center_point=nil
				if( @state == 0 )
					@ip.pick view, x, y
					if @ip != @ip1
						view.tooltip = @ip.tooltip
						set_special_point(x,y,view,@ip) #if !view.inference_locked?
						view.invalidate
						@ip1.copy! @ip
					end
					if @line_style==1 || @alt_press
						@list_edges= get_point_from_edge(@edges,[@ip1.position,@ip1.position])
						@list_faces= get_point_from_face(@faces,[@ip1.position,@ip1.position])
						@list_curves= get_point_from_curve(@curves,[@ip1.position,@ip1.position])

						@list_edges.each_index{|i| @list_edges[i][1].each_index{|j| @list_edges[i][1][j].z=@ip1.position.z}}
						@list_curves.each_index{|i| @list_curves[i][1].each_index{|j| @list_curves[i][1][j].z=@ip1.position.z}}
						@list_faces.each_index{|i| @list_faces[i][1].each_index{|j| @list_faces[i][1][j].z=@ip1.position.z}}
					end
					set_status_text(view)
				elsif !@nodraw
					@ip.pick view, x, y, @ip1
					view.tooltip = @ip.tooltip if( @ip.valid? )
					set_special_point(x,y,view,@ip) #if !view.inference_locked?
					if @shift_press && @p_shift && @ip!=@ip1 && @ip2!=@ip1
						p2_line=@ip.position.project_to_line([@ip1.position,@p_shift])
						@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
					elsif @ip_lock && @ip!=@ip1 && @ip2!=@ip1
						p2_line=@ip.position.project_to_line([@ip1.position,@ip_lock])
						@ip2.copy!(p2_line ? Sketchup::InputPoint.new(p2_line) : @ip)
					else
						@ip2.copy! @ip
					end
					@list_edges= get_point_from_edge(@edges,[@ip1.position,@ip2.position])
					@list_faces= get_point_from_face(@faces,[@ip1.position,@ip2.position])
					@list_curves= get_point_from_curve(@curves,[@ip1.position,@ip2.position])

					length = @ip1.position.distance(@ip2.position)
					set_status_text(view,length)
				end
			end
			def onLButtonDown(flags, x, y, view)
				if @state == 0
					if @cancel
						@cancel.abort_operation
						@cancel=nil
						#view.model.select_tool nil
					end
					if( @ip1.valid? )
						@list_edges=[]
						@list_curves=[]
						@list_faces=[]
						if @line_style==1 || @alt_press
							p=@ip1.position
							p.z=0
							self.create_geometry([p,@ip1.position])
							view.model.select_tool nil
							#self.reset(view)
							#@nodraw=true
							#set_status_text(view)
						else
							@state = 1
						end
					end
				elsif @state == 1
					if @ip2.valid?
						self.create_geometry()
						@nodraw=true
						set_status_text(view)
						#self.reset(view)
						#view.model.select_tool nil
					end
				end
				set_status_text(view)
			end
			def onKeyDown(key, repeat, flags, view)
				if( key == CONSTRAIN_MODIFIER_KEY )
					@shift_press =true
					if( view.inference_locked? )
						view.lock_inference
					elsif( @state == 0 && @ip1.valid? )
						view.lock_inference @ip1
					elsif( @state == 1 && @ip2.valid? )
						@p_shift=@ip.position
						view.lock_inference @ip2, @ip1
					end
				elsif key==VK_UP
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p=  @ip1.position
						p1=p + @edit_transform.zaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==VK_LEFT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p1=@ip1.position + @edit_transform.yaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key==VK_RIGHT
					view.lock_inference if view.inference_locked?
					if @ip.valid? && @state>0
						p1=@ip1.position + @edit_transform.xaxis
						ip1 = Sketchup::InputPoint.new(p1)
						view.lock_inference ip1,@ip
						@ip_lock=ip1.position
						@ip2.copy! ip1
					end
				elsif key == ALT_MODIFIER_KEY
					@alt_press= true
				end
			end
			def onReturn(view)
				if @ip2.valid?
					self.create_geometry()
					@nodraw=true
					set_status_text(view)
				elsif @last_tool
					self.create_geometry(@last_tool)
					@nodraw=true
					set_status_text(view)
				end
			end
			def onKeyUp(key, repeat, flags, view)
				if( key == CONSTRAIN_MODIFIER_KEY )
					@shift_press=nil
					@p_shift=nil
					view.lock_inference
				elsif key == ALT_MODIFIER_KEY
					@alt_press=nil
				end
			end
			def onUserText(text, view)
				if @state==0
					value=cal(text)
					return if !value || value==0
					vec=Z_AXIS.clone
					vec.length=value
					pt2 = @ip1.position + vec
					@ip2.copy! Sketchup::InputPoint.new(pt2)
					self.create_geometry()
					@nodraw=true
					set_status_text(view)
					#view.model.select_tool nil
					#@state=1
				elsif @state == 1
					vec = @ip2.position - @ip1.position
					return if vec.length==0
					value=cal(text)
					return if !value || value==0
					vec.length = value
					pt2 = @ip1.position + vec
					set_status_text(view,value)
					@ip2.copy! Sketchup::InputPoint.new(pt2)
					self.create_geometry()
					@nodraw=true
					set_status_text(view)
					#view.model.select_tool nil
				end
			end
			def draw(view)
				if @state==0  &&  !@nodraw
					if @ip
						if @ip.valid?
							@ip.draw(view)
						end
						if @face || @edge
							view.draw_points @ip.position, 9*@scale_factor, 1, "red"
						elsif @center_point
							view.draw_points @ip.position, 9*@scale_factor, 1, "red"
						end
					end
					if @line_style==1 || @alt_press
						view.drawing_color ="blue"
						line_style=["","-"]
						view.line_width=2
						view.line_stipple = line_style[0]
						(@list_edges+@list_faces+@list_curves).compact.each{|list| view.draw_polyline(list[1]) if list[1]}
						view.line_width=1
						view.line_stipple = line_style[1]
						(@list_edges+@list_faces+@list_curves).compact.each{|list|
							list[0].each_index{|i|
								view.draw_line(list[0][i],list[1][i]) if list[0][i]
							} if list[0]
						}
					end
				elsif @state>0 && !@nodraw
					if @ip1.valid?
						@ip1.draw(view)
					end
					if @ip
						if @ip.valid?
							@ip.draw(view)
						end
						if @face || @edge
							view.draw_points @ip.position, 9*@scale_factor, 1, "red"
						elsif @center_point
							view.draw_points @ip.position, 9*@scale_factor, 1, "red"
						end
					end
					if @ip2.valid? && @line_style<1
						if @p_shift ||((@face || @edge) && @ip_lock)
							p=@p_shift
							p = @ip_lock if (@face || @edge) && @ip_lock
							p2_line=@ip.position.project_to_line([@ip1.position,p])
							view.draw_points p2_line, 9*@scale_factor, 2, "red"  if p2_line
							@ip2.copy! Sketchup::InputPoint.new(p2_line)
							view.line_stipple = "."
							view.drawing_color ="blue"
							view.draw_line(@ip.position, @ip2.position)
							view.line_stipple = ""
							view.drawing_color ="black"
						end
						@ip2.draw(view) if @ip2.display?
						view.set_color_from_line(@ip1, @ip2)
						line_style=["","-"]
						view.line_width=2
						view.line_stipple = line_style[0]
						view.draw_line(@ip1.position, @ip2.position)
						(@list_edges+@list_faces+@list_curves).compact.each{|list| view.draw_polyline list[1]}
						view.line_width=1
						view.line_stipple = line_style[1]
						(@list_edges+@list_faces+@list_curves).compact.each{|list|
							list[0].each_index{|i|
								view.draw_line(list[0][i],list[1][i])
							}
						}
						view.draw_line(@ip1.position, @ip2.position)
					end
				end
			end
			def onCancel(flag, view)
				if @cancel
					if flag==2
						@cancel.commit_operation
					else
						@cancel.abort_operation
					end
					@cancel=nil
				end
				self.reset(view)
			end
			def reset(view=nil)
				@state = 0
				@picked_element=[]
				@ip1.clear
				@ip2.clear
				@center=[]
				@ip_lock=nil
				set_status_text(view)
			end
			def set_status_text(view=nil,value=nil)
				text=@list_type[@line_style] + " | "
				Sketchup::set_status_text @strings["Length"], SB_VCB_LABEL
				if @state == 0
					if @line_style==0
						text += @strings["Select First Point"]
					else
						text += @strings["Select Point"]
					end
				elsif @state == 1
					text += @strings["Select Second Point"]
				end
				if IS_WIN
					text += " | " +  @strings["Hold Alt and Click for Set Height"]
				else
					text += " | " +  @strings["Hold Command and Click for Set Height"]
				end
				if value
					Sketchup::set_status_text value.to_s, SB_VCB_VALUE
				end
				Sketchup.status_text = text
				if view
					view.invalidate
					#view.lock_inference
				end
			end
			def cal(input)
				if op=input[/[\+\-\/\*]/]
					begin
					if input[/[\+\-]/]
						index= input.rindex(/[\+\-]/)
					elsif input[/[\*\/]/]
						index= input.rindex(/[\*\/]/)
					end
					if index
						if index>0
							s=[input[0..(index-1)],input[(index+1)..-1]]
						else
							s=["0",input[(index+1)..-1]]
						end
						op=input[index]
						if op=="+"
							value= cal(s[0]).to_s.to_f + cal(s[1]).to_s.to_f
						elsif op=="-"
							value= cal(s[0]).to_s.to_f - cal(s[1]).to_s.to_f
						elsif op=="*"
							value= cal(s[0]).to_s.to_f * cal(s[1]).to_s.to_f
						elsif op=="/"
							value= cal(s[0]).to_s.to_f / cal(s[1]).to_s.to_f
						end
					end
					rescue
						value=parse_length(input)
					end
				else
					value = parse_length(input)
				end
				return parse_length(value)
			end
			def parse_length(input)
				(return 0) if !input
				input=input.to_s
				if !Sketchup.parse_length("1.1")
					input.gsub!(/\./, ',')
				else
					input.gsub!(/\,/, '.')
				end
				begin
					value=input.to_l
				rescue
					value=input.to_f
				end
				return value
			end
			def get_point_from_face(faces,ps)
				hidden=Sketchup.active_model.rendering_options['DrawHidden']
				result=[]
				vec=ps[1]-ps[0]
				faces.each{|face|
					face.loops.each{|loop|
						vers=loop.vertices.clone
						vers<<vers[0]
						points=vers.collect{|p| p.position}
						points1=points.collect{|p| p+vec}
						result<<[points,points1,face,vers,(loop == face.outer_loop)]
					}
				}
				return result
			end
			def get_point_from_curve(cuvers,ps)
				hidden=Sketchup.active_model.rendering_options['DrawHidden']
				result=[]
				vec=ps[1]-ps[0]
				cuvers.each{|curve|
					 # if hidden || (!hidden && !curve.edges.any?{|e| e.soft? || e.hidden?})
					if curve && (curve.is_a?Sketchup::Curve) && curve.valid?
						points=curve.vertices.collect{|p| p.position}
						points1=points.collect{|p| p+vec}
						result<<[points,points1,curve,curve.vertices]
					end
				}
				return result
			end
			def get_point_from_edge(edges,ps)
				hidden=Sketchup.active_model.rendering_options['DrawHidden']
				vec=ps[1]-ps[0]
				result=[]
				@list_continous_edges.each{|list|
					list.each{|li|
						ps=[]
						points=li.collect{|p| p.position}
						points1=points.collect{|p| p+vec}
						v1=li[0]
						v2=li[1]
						edge=((v1.edges & v2.edges) & edges)[0]
						# (0..li.length-2).each{|i|
							# v1=li[i]
							# v2=li[i+1]
							# edge=((v1.edges & v2.edges) & edges)[0]
							# next if !edge
							# next if ps[0].on_line?(edge.line) && ps[1].on_line?(edge.line)
							# if hidden || (!hidden && !(edge.soft? || edge.hidden?))
								# points=edge.vertices.collect{|p| p.position}
								# points=[v1.position,v2.position]
								# points1=points.collect{|p| p+vec}
								# result<<[points,points1,edge,edge.vertices]
							# end
						# }
						result<<[points,points1,edge,li]
					}
				}
				return result
			end
			def get_point_from_edge1(edges,ps)
				hidden=Sketchup.active_model.rendering_options['DrawHidden']
				vec=ps[1]-ps[0]
				result=[]
				@list_continous_edges.each{|list|
					list.each{|li|
						(0..li.length-2).each{|i|
							v1=li[i]
							v2=li[i+1]
							edge=((v1.edges & v2.edges) & edges)[0]
							next if !edge
							next if ps[0].on_line?(edge.line) && ps[1].on_line?(edge.line)
							if hidden || (!hidden && !(edge.soft? || edge.hidden?))
								# points=edge.vertices.collect{|p| p.position}
								points=[v1.position,v2.position]
								points1=points.collect{|p| p+vec}
								result<<[points,points1,edge,edge.vertices]
							end
						}
					}
				}
				return result
			end
			def make_face(points)
				if points
					@list_edges= get_point_from_edge(@edges,points)
					@list_faces= get_point_from_face(@faces,points)
					@list_curves=get_point_from_curve(@curves,points)
				end
				@group=Sketchup.active_model.active_entities.add_group
				#entities=Sketchup.active_model.active_entities
				entities=@group.entities
				faces_edge_from_face=[]
				faces_from_face=[]
				faces_edge_from_edge=[]
				faces_from_curve=[]
				if @line_style==1 || @alt_press
					@list_edges.each_index{|i| @list_edges[i][1].each_index{|j| @list_edges[i][1][j].z=points[1].z}}
					@list_curves.each_index{|i| @list_curves[i][1].each_index{|j| @list_curves[i][1][j].z=points[1].z}}
					@list_faces.each_index{|i| @list_faces[i][1].each_index{|j| @list_faces[i][1][j].z=points[1].z}}
				end
				@list_edges.each{|list|
					faces_edge_from_edge+=S4U::S4u_Linetoface.extrude_points_points(entities,list,false,@hash_face_of_edges)
				}
				ent=entities.to_a
				@list_faces.each{|list|
					face=S4U::S4u_Linetoface.extrude_points_points(entities,list,false,[])
					# faces_from_face+=face
				}
				@list_curves.each{|list|
					g=entities.add_group
					g.entities.add_curve list[1]
					g.explode if g && g.valid?
				}
			end
			def create_geometry(points=[@ip1.position, @ip2.position])
				model=Sketchup.active_model
				entities=model.active_entities
				model.selection.clear
				@cancel.abort_operation if @cancel
				model.start_operation(TOOL,true,false,false)
				@cancel=model
				@last_tool=points.clone if points && points[0].distance(points[1])>0.0
				make_face(points)
				#model.commit_operation
				rescue =>error
				p error
				p error.backtrace.join('\n')
				model.abort_operation
			end
		end
		def self.delete_hole(faces)
				faces1=faces.clone
				fd=[]
				i=0
				length=faces.length
				while (faces.length > 0) &&(i< length)
					f=faces.shift
					(f.loops - [f.outer_loop]).each{|l|
						fs=l.edges[0].faces
						l.edges.each{|e| fs= e.faces & fs }
						(fs-[f]).each{|fa|
							if fa.valid? && fa.loops.length>1
								faces = faces - [fa]
								fd<< fa
							end
						}
					}
					i+=1
				end
				fd.uniq!
				faces=faces1-fd
				fd1=[]
				i=0
				length=faces.length
				while (faces.length > 0) && (i< length)
					f=faces.shift
					(f.loops - [f.outer_loop]).each{|l|
						fs=l.edges[0].faces
						l.edges.each{|e| fs = e.faces & fs}
						(fs-[f]).each{|fa|
							 if  fa.valid?
								 faces = faces - [fa]
								 fd1<< fa
							 end
						 }
					}
					i+=1
				end
				fd1.uniq!
				faces1 -= (fd1 + fd)
				(fd1 + fd).each{|f| f.erase! if f && f.valid?}
				return faces1
				rescue => error
				p error
				p error.backtrace.join('\n')
				return nil
			end
		def self.extrude_points_points(entities,list,weld=false,hash_face)
				hidden=Sketchup.active_model.rendering_options['DrawHidden']
				fs=[]
				ed,ed1=list[0],list[1]
				g=entities.add_group
				(0..ed.length-2).each{|i|
					if !ed1[i].on_line?([ed[i],ed[i+1]]) || !ed1[i+1].on_line?([ed[i],ed[i+1]])
						v1=list[3][i]
						v2=list[3][i+1]
						edge= (v1.edges & v2.edges)[0]
						if edge && ( hidden || (!hidden && !(edge.hidden? || edge.soft? )))
							soft1=edge.soft?
							smooth1=edge.smooth?
							hidden1=edge.hidden?
							eds=ed[i],ed[i+1],ed1[i+1],ed1[i]
							if ed[i+1]==ed1[i+1] || (ed[i+1].distance(ed1[i+1]).to_f <=0.0001)
								eds.delete_at(2)
							end
							if ed[i]==ed1[i] || (ed[i].distance(ed1[i]).to_f <=0.00001)
								eds.delete_at(3)
							end
							if eds.length>2
								v1_curve_interior=v1.curve_interior?
								v2_curve_interior=v2.curve_interior?
								f=g.entities.add_face eds
								if f
									vec=ed[i+1]-ed[i]
									vec1=ed1[i]-ed[i]
									(vec1=ed1[i+1]-ed[i+1]) if vec1.length==0
									n=vec.cross vec1
									if n.length!=0 && (n.samedirection?(f.normal.reverse)  ||  n.angle_between(f.normal.reverse)<Math::PI*0.5)
										f.reverse!
									end
									if v1_curve_interior
										v=f.vertices.find{|v| v.position.distance(ed[i]).to_f <=0.000001}
										if v
											e=(v.edges & f.edges).find{|e| e.other_vertex(v).position.distance(ed[i+1])>=0.000001}
											if e
												e.soft=true
												e.smooth=true
											end
											e=(v.edges & f.edges).find{|e| e.other_vertex(v).position.distance(ed1[i])<=0.000001}
											if e
												e.soft=true
												e.smooth=true
											end

										end
									end
									if v2_curve_interior
										v=f.vertices.find{|v| v.position.distance(ed[i+1]).to_f <=0.000001}
										if v
											e=(v.edges & f.edges).find{|e| e.other_vertex(v).position.distance(ed[i])>=0.000001}
											if e
												e.soft=true
												e.smooth=true
											end
											e=(v.edges & f.edges).find{|e| e.other_vertex(v).position.distance(ed1[i+1])<=0.000001}
											if e
												e.soft=true
												e.smooth=true
											end
										end
									end
									v3=f.vertices.find{|v| v.position.distance(ed1[i+1]).to_f <=0.000001}
									v4=f.vertices.find{|v| v.position.distance(ed1[i]).to_f <=0.000001}
									if v3 && v4
										edge1= (v3.edges & v4.edges)[0]
										if edge1
											edge1.soft=soft1
											edge1.smooth=smooth1
											edge1.hidden=hidden1
										end
									end
									fs<<f
								end
							end

						end
					end
				}
				if list[2].class==Sketchup::Face
					face=list[2]
					if fs[0] && check_direction_face(face,[ed[0],ed[1]])[1].samedirection?(fs[0].normal) #&& list[4]
						fs.each{|f| f.reverse!}
					end
				elsif list[0][0]==list[0][-1]
					if fs[0] && check_direction_edge(list[0],[ed[0],ed[1]])[1].samedirection?(fs[0].normal)
						 fs.each{|f| f.reverse! }
					end
				else
					if list[2].class==Sketchup::Edge && (list[0][0]!=list[0][-1])
						if fs[0] && check_direction_edge(list[0],[ed[0],ed[1]])[1].samedirection?(fs[0].normal)
							 fs.each{|f| f.reverse! }
						end
						if list[2].faces.length>0
							fs1=list[2].faces
							fs1.select!{|f1| hash_face[f1] }
							fs1.compact!
							f=fs[0]
							stop=false
							fs1.each{|face|
								if f.normal.samedirection?(face.normal.reverse) || f.normal.angle_between(face.normal.reverse)<Math::PI*0.1
									fs.each{|f1| f1.reverse! }
									stop=true
									break
								end
							}
							if !stop
								face=fs1.max_by{|f1| hash_face[f1] }
								if face
									vy=check_direction_face(face,[ed[0],ed[1]])[1]
									if vy.samedirection?(f.normal)
										fs.each{|f1| f1.reverse! }
									end
								end
							end
						else
							fs.each{|f|
								if f.normal.samedirection?(Z_AXIS.reverse) ||  f.normal.angle_between(Z_AXIS.reverse)<Math::PI*0.5
									 f.reverse!
								end
							}
						end
					end
				end
				if list[2].class==Sketchup::Face
					face=list[2]
					mesh=face.mesh(7)
					t=list[1][0]-list[0][0]
					trans=Geom::Transformation.translation(t)
					if trans
						mesh1=mesh.transform! trans
						g1=entities.add_group
						g1.entities.add_faces_from_mesh mesh1
						face1=g1.entities.select{|e| e && e.valid? && (e.class==Sketchup::Face) }
						if face1.length >0
							face1.each{|f|
								p = list[0][0].project_to_plane f.plane
								v=list[0][0]-p
								if v.length>0 && f.normal.samedirection?(v)
									f.reverse!
								end
								if f.normal.samedirection?(face.normal)
									 face.reverse!
								end
							}
						end
						self.clean_edge(g1.entities,true)
						g1.explode if g1 && g1.valid?
					end
				end
				fs=g.explode if g && g.valid?
				if weld
					g=entities.add_group
					g.entities.add_curve ed1
					g.explode if g && g.valid?
				end
				return fs.select{|e| e && e.valid? && e.is_a?(Sketchup::Face)}
		end
		def self.get_ocs(z_axis)
			  if (z_axis.x.abs < (1.0/64.0)) && (z_axis.y.abs < (1.0/64.0))
				x_axis = Y_AXIS*z_axis
			  else
				x_axis = Z_AXIS*z_axis
			  end
			  y_axis = z_axis*x_axis
			  return Geom::Transformation.axes(ORIGIN, x_axis, y_axis, z_axis)
			  rescue => error
			  return Geom::Transformation.new(z_axis)
		end
		def self.normal_vertices(ver)
			return Z_AXIS if ver.length<2
			vx=nil;vy=nil;line=nil
			(0..ver.length-2).each{|i|
				vx=ver[i].vector_to ver[i+1]
				line=[ver[i],ver[i+1]]
				break if vx.length!=0
			}
			if line
				(0..ver.length-1).each{|i|
					v2=ver[i].project_to_line line
					vy=v2.vector_to ver[i]
					break if vy.length!=0
				}
				v= (vx.cross vy)
				return v if vx && vy && v.length!=0
			end
			trans=get_ocs(ver[1]-ver[0])
			return trans.yaxis
			if ver[0].z==ver[1].z
				return Z_AXIS.clone
			elsif ver[0].y==ver[1].y
				return Y_AXIS.clone
			else
				return X_AXIS.clone
			end
		end
		def self.point_in_polygon_2D(p,ps,check_border)
			if ps.length>2
				vz= normal_vertices ps
				vz=Z_AXIS if !vz || vz.length==0
				t=Geom::Transformation.new vz
				ps1=ps.collect{|pt| pt.transform t.inverse}
				p1=p.transform t.inverse
				return  Geom::point_in_polygon_2D(p1,ps1,check_border)
			elsif ps.length==2
				if p.on_line?(ps)
					if check_border
						return (p.distance(ps[0])<=ps[0].distance(ps[1]) && p.distance(ps[1])<=ps[0].distance(ps[1]) )
					else
						return (p.distance(ps[0])<ps[0].distance(ps[1]) && p.distance(ps[1])<ps[0].distance(ps[1]) )
					end
				else
					return false
				end
			else
				if check_border
					return (p.distance(ps[0])<=0.00001)
				else
					return (p.distance(ps[0])>0)
				end
			end
		end
		def self.check_direction_edge(ps,points)
			view=Sketchup.active_model.active_view
			if (points.length>1) && (points[1]!=points[0])
				v=points[1]-points[0]
				vz=normal_vertices ps
				vy=vz.cross v
				if vy.length!=0
					vy.length=points[0].distance(points[1])*0.1
					pm=Geom::Point3d.new (points[0].x+points[1].x)*0.5,(points[0].y+points[1].y)*0.5,(points[0].z+points[1].z)*0.5
					p = pm + vy
					if ps.length>1 && (ps[0].distance(ps[-1]).to_f <= 0.00001)
						if self.point_in_polygon_2D(p,ps,true)
							return [-1,vy]
						else

							return [1,vy.reverse]
						end
					end
					# direction=1
					# vec=nil
					# if ps.length==2
						# vp=points[1]-points[0]
						# v=ps[1]-ps[0]
						# vec=v.cross vp
						# vecz=S4U::S4u_Offset.normal_vertices(ps)
						# direction= (vec.length!=0 && vec.samedirection?(vecz)) ? 1 : -1
					# else
						# line1=[view.screen_coords(points[0]),view.screen_coords(points[1])]
						# v=line1[1]-line1[0]
						# p1=view.screen_coords(points[1])
						# p=p1.project_to_line line1
						# v1=p1-p
						# vec=v1.cross v
						# vecz=S4U::S4u_Offset.normal_vertices(ps)
						# direction= (vec.length!=0 && vec.samedirection?(vecz)) ? 1 : -1
					# end
					# return [direction,vec]
					p1=p.project_to_line ps[0..1]
					v1=p-p1
					if v1.length!=0 && v1.samedirection?(vy)
						return [1,v1]
					else
						return [-1,v1.reverse]
					end
				end
			end
			return [0,Z_AXIS.clone]
		end
		def self.check_direction_face(face,points1)
			points=points1.collect{|p| p.project_to_plane(face.plane)}
			if (points.length>1) && (points[1]!=points[0]) && face
				v=points[1]-points[0]
				vy=face.normal.cross v
				if vy.length!=0
					vy.length=points[0].distance(points[1])*0.01
					pm=Geom::Point3d.new (points[0].x+points[1].x)*0.5,(points[0].y+points[1].y)*0.5,(points[0].z+points[1].z)*0.5
					p = pm + vy
					if face.classify_point(p)==1
						return [-1,vy]
					else
						return [1,vy.reverse]
					end
				end
			end
			return [0,face.normal]
		end
		def self.clean_edge(entities,delete=false,ang=0.15)
			edge_selected=[]
			edge_selected1=[]
			faces=entities.grep(Sketchup::Face)
			edges=entities.grep(Sketchup::Edge)
			edges+= faces.collect{|f| f.edges}.flatten.uniq
			edges.each{|e|
				if e.faces.length==0 ||
				(e.faces.length==2 &&
				#e.faces[0].material==e.faces[1].material && e.faces[0].back_material==e.faces[1].back_material &&
				#e.faces[0].normal.samedirection?( e.faces[1].normal))
				(e.faces[0].normal.angle_between(e.faces[1].normal) <= 0.0000001))
					edge_selected << e
				elsif e && e.valid? && (e.faces.length==2) &&
					(e.faces[0].normal.angle_between( e.faces[1].normal) <= Math::PI*ang )
						edge_selected1 << e
				end
			}
			if delete
				edge_selected.each{|e| e.erase! if e.valid? }
			else
				edge_selected1+=edge_selected
			end
			(edge_selected1).compact.each{|e|
				(e.smooth=e.soft=true) if  e && e.valid?
			}
		end
		def self.find_weld(edges,check_hide_edge=true)
			ver_list_connected=[]
			if edges.length>0
				edges=edges.select{|e| !(e.hidden? || e.soft?)}.uniq if check_hide_edge
				edge_list_connected= find_all_connected edges
				ver_list_connected=edge_list_connected.collect{|eds| find_vertices(eds)}
			end
			return ver_list_connected
		end
		def self.find_all_connected(edges)
			edges1=edges.clone
			i=0;list=[]
			while (edges1.length>0) && (i<edges.length)
				ls=[]
				if f=edges1.find{|e| e.vertices[0].edges.length==1 || e.vertices[1].edges.length==1}
					ls<<f
					edges1.delete f
				else
					ls<<edges1.shift
				end
				allconnect= (ls.last.all_connected) & edges
				ls+=allconnect
				ls.uniq!
				list<< ls
				edges1-=ls
				i+=1
			end
			return list
			rescue =>error
			p error
			p error.backtrace.join('\n')
			return edges
		end
		def self.find_vertices(edges_new1)
			edges_new=edges_new1.clone
			i=0
			list=[]
			list1=[]
			while (edges_new.length>0) && (i<edges_new1.length)
				vers=edges_new.flat_map{|e| e.vertices}.uniq
				ls=[]
				if f=vers.find{|v| (l=(v.edges & edges_new1).length )&& l && ((l==1) ||(l>2) )}
					ls<<f
					vers.delete f
				elsif f=vers.find{|v| (l=(v.edges ).length )&& l && ((l==1) || (l>2))}
					ls<<f
					vers.delete f
				elsif f=vers.find{|v| (v.edges & edges_new).length==1}
					ls<<f
					vers.delete f
				else
					ls<< vers.shift
				end
				length=vers.length
				j=0
				while (edges_new.length>0) && (j<length) && (vers.length>0)
					j+=1
					vs=ls.last.edges.flat_map{|e| e.vertices} & vers
					lc=vs.find{|l|
							l.common_edge(ls.last) &&
							((l.edges & ls.last.edges) & edges_new).length==1
						}
					if lc
						led=(lc.edges & ls.last.edges) & edges_new
						edges_new-=led
						ls<< lc
						vers.delete(lc)
						break if ((lc.edges & edges_new1).length>2)
					end
				end
				if ls.length<=2
					list1 << ls
				else
					list << ls
				end
				i+=(ls.length-1)
			end

			j=0
			length=list1.length
			while list1.length>0 && j<length
					ls1=list1[0]
					list.each_index{|i|
						if ((ls1 & list[i]).length==2) && (((list[i][0].edges & list[i][-1].edges) & edges_new1).length==1)
							if (list[i][0].edges).length>(list[i][-1].edges).length
								list[i]<<list[i][0]
								list1.delete_at(0)
								break
							elsif (list[i][0].edges ).length<(list[i][-1].edges ).length
								list[i].unshift list[i][-1]
								list1.delete_at(0)
								break
							elsif (list[i][0].edges & edges_new1).length>(list[i][-1].edges & edges_new1).length
								list[i]<<list[i][0]
								list1.delete_at(0)
								break
							elsif (list[i][0].edges & edges_new1).length<(list[i][-1].edges & edges_new1).length
								list[i].unshift list[i][-1]
								list1.delete_at(0)
								break

							elsif ls1[0]==list[i][0] && ls1[-1]==list[i][-1] #&& ls1[0].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2 &&
								#ls1[1].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2
								list[i]<<ls1[0]
								list1.delete_at(0)
								break
							elsif ls1[0]==list[i][-1] && ls1[-1]==list[i][0] #&& ls1[0].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2 &&
								#ls1[1].edges.select{|e|!(e.soft? || e.hidden?)}.length<=2
								list[i]<<ls1[-1]
								list1.delete_at(0)
								break
							end
						end
					}
				j+=1
			end
			return (list + list1)
		end
		def self.main(mode)
			if !@setting
				@setting=[0,0]
			end
			Sketchup.active_model.select_tool LineToFace.new
		end
	end
end